"""
Author:         David Szybala
Website:        www.davidszybala.com
Alt Website:    www.twistedgerbil.com
Program:        material_list.py
Date:           11/03/14
Description:    This script makes searching/graphing/assigning materials
                fast and easy
Installation:   Place this file in your maya scripts directory
                (e.g. ..\My Documents\maya\scripts) then run the following
                python commands:
                    import material_list
                    material_list.main()
"""
import maya.cmds as cmds
import maya.mel as mel


class MaterialList(object):
    def __init__(self):
        self.materials_root = []
        self.materials_all = []
        self.materials_offsets = []
        self.materials_filtered = []
        self.materials_offsets_filtered = []
        self.material_classes = []

        self.material_frame = None
        self.material_list = None
        self.filter_text = None
        self.include_unused_checkbox = None
        self.include_sub_materials_checkbox = None
        self.graph_input_radio = None
        self.graph_output_radio = None
        self.status_label = None

    @staticmethod
    def is_hypershade_open():
        """Returns True if the hypershade panel is open"""
        all_windows = cmds.lsUI(windows=True)

        if 'hyperShadePanel1Window' in all_windows:
            return True
        else:
            return False

    @staticmethod
    def is_material_node(current_node):
        """Returns True if the 'current_node' is a shader class"""
        is_shader = False

        if current_node:

            node_class = cmds.getClassification(cmds.nodeType(current_node))
            if 'shader/surface' in node_class[0]:
                is_shader = True

        return is_shader

    @staticmethod
    def get_root_materials(include_unused):
        """Returns a sorted list of root materials in the current scene"""
        root_materials = []

        for current in cmds.ls(type='shadingEngine', long=True):
            shader = cmds.listConnections(current + '.surfaceShader',
                                          source=True, destination=False)
            if not shader:
                print('WARNING: Shading group ' + current + ' has no surface shader')
                continue

            if include_unused:
                if shader[0] not in root_materials:
                    root_materials.append(shader[0])
            else:
                if cmds.sets(current, query=True):
                    if shader[0] not in root_materials:
                        root_materials.append(shader[0])

        root_materials.sort()

        return root_materials

    def collect_sub_materials(self, parent_material, offset_value):
        """Recursive function. Collects all the sub materials from the argument
        'parent_material' """
        node_connections = cmds.listConnections(parent_material, source=True, destination=False)
        name_offset = ''

        if node_connections:
            node_connections = set(node_connections)
            node_connections = list(node_connections)
            node_connections.sort()

            for current in node_connections:
                if self.is_material_node(current):

                    # CRITICAL! If material is already in list then skip!
                    if current in self.materials_all:
                        continue

                    self.materials_all.append(current)

                    for i in range(0, offset_value):
                        name_offset += '   '

                    self.materials_offsets.append(name_offset)

                    self.collect_sub_materials(current, offset_value + 1)

    def filter_scene_materials(self, filter_text):
        """Filters the current scene materials by name. NOTE: Sub materials
        are NOT affected"""
        collect_subs = False

        if filter_text:

            # CRITICAL! Clear previous data!
            self.materials_filtered = []
            self.materials_offsets_filtered = []

            for i in range(0, len(self.materials_offsets)):
                if not self.materials_offsets[i]:
                    if filter_text in self.materials_all[i]:
                        self.materials_filtered.append(self.materials_all[i])
                        self.materials_offsets_filtered.append(self.materials_offsets[i])
                        collect_subs = True
                    else:
                        collect_subs = False
                else:
                    if collect_subs:
                        self.materials_filtered.append(self.materials_all[i])
                        self.materials_offsets_filtered.append(self.materials_offsets[i])

        else:
            self.materials_filtered = self.materials_all
            self.materials_offsets_filtered = self.materials_offsets

    def collect_scene_materials(self, filter_text, include_unused, include_subs):
        """Stores scene material names, offsets and a filtered list"""
        self.materials_root = self.get_root_materials(include_unused)

        # CRITICAL! Clear previous data!
        self.materials_all = []
        self.materials_offsets = []

        for current in self.materials_root:
            self.materials_all.append(current)
            self.materials_offsets.append('')

            if include_subs:
                self.collect_sub_materials(current, 1)

        self.filter_scene_materials(filter_text)

    @staticmethod
    def get_shading_group(current_object):
        """Returns a list of shading groups of the 'current_object' """
        shading_groups = []

        if cmds.nodeType(current_object) == 'transform':
            children = cmds.listRelatives(current_object, shapes=True,
                                          fullPath=True, noIntermediate=True)
            if children:
                for current in children:
                    connections = cmds.listConnections(current, type='shadingEngine',
                                                       source=False, destination=True)
                    if connections:
                        shading_groups.extend(connections)
        else:
            connections = cmds.listConnections(current_object, type='shadingEngine',
                                               source=False, destination=True)
            if connections:
                shading_groups.extend(connections)

        return shading_groups

    def refresh_material_list(self, *args):
        """Refreshes the material list names"""
        include_unused = cmds.checkBox(self.include_unused_checkbox, query=True, value=True)
        include_subs = cmds.checkBox(self.include_sub_materials_checkbox, query=True, value=True)
        filter_text = cmds.textFieldButtonGrp(self.filter_text, query=True, text=True)

        self.collect_scene_materials(filter_text, include_unused, include_subs)

        cmds.textScrollList(self.material_list, edit=True, removeAll=True)

        if len(self.materials_filtered) > 0:
            display_names = [self.materials_offsets_filtered[i] + self.materials_filtered[i]
                             for i in range(0, len(self.materials_filtered))]
            cmds.textScrollList(self.material_list, edit=True, append=display_names)

        cmds.frameLayout(self.material_frame, edit=True, label=('Materials: ' + str(len(self.materials_root))))
        cmds.text(self.status_label, edit=True, label='INFO: Material list updated')

    def filter_text_changed(self, *args):
        """Updates the material list based on the new filter"""
        filter_text = cmds.textFieldButtonGrp(self.filter_text, query=True, text=True)

        self.filter_scene_materials(filter_text)

        cmds.textScrollList(self.material_list, edit=True, removeAll=True)

        if len(self.materials_filtered) > 0:
            display_names = [self.materials_offsets_filtered[i] + self.materials_filtered[i]
                             for i in range(0, len(self.materials_filtered))]
            cmds.textScrollList(self.material_list, edit=True, append=display_names)

        cmds.text(self.status_label, edit=True, label='INFO: Material list filtered')

    def graph_material_button_pressed(self, *args):
        """Graphs the selected materials in the hypershade"""
        user_selection = cmds.textScrollList(self.material_list, query=True, selectIndexedItem=True)

        if not user_selection:
            cmds.text(self.status_label, edit=True, label='ERROR: No materials selected')
            return False

        if not self.is_hypershade_open():
            cmds.text(self.status_label, edit=True, label='ERROR: Hypershade must already be open')
            return False

        select_materials = [self.materials_filtered[index - 1] for index in user_selection]
        cmds.select(select_materials)

        if cmds.radioButton(self.graph_input_radio, query=True, select=True):
            mel.eval('hyperShadePanelGraphCommand("hyperShadePanel1", "showUpstream");')
        elif cmds.radioButton(self.graph_output_radio, query=True, select=True):
            mel.eval('hyperShadePanelGraphCommand("hyperShadePanel1", "showDownstream");')
        else:
            mel.eval('hyperShadePanelGraphCommand("hyperShadePanel1", "showUpAndDownstream");')

        cmds.text(self.status_label, edit=True,
                  label=('INFO: ' + str(len(user_selection)) + ' materials graphed'))

    def clear_graph_button_pressed(self, *args):
        """Clear graph event handler. Clears the hypershade graph"""
        if self.is_hypershade_open():
            mel.eval('hyperShadePanelGraphCommand("hyperShadePanel1", "clearGraph");')
            cmds.text(self.status_label, edit=True, label='INFO: Hypershade graph cleared')
        else:
            cmds.text(self.status_label, edit=True, label='ERROR: Hypershade is not open')

    def open_hypershade_button_pressed(self, *args):
        """Open hypershade window event handler. Opens the hypershade window"""
        if not self.is_hypershade_open():
            cmds.HypershadeWindow()
            cmds.text(self.status_label, edit=True, label='INFO: Hypershade window is now open')
        else:
            cmds.text(self.status_label, edit=True, label='ERROR: Hypershade is already open')

    def select_objects_by_material_button_pressed(self, *args):
        """Select objects by material event handler. Selects all objects with
        the selected material assigned"""
        user_selection = cmds.textScrollList(self.material_list, query=True, selectIndexedItem=True)

        if not user_selection:
            cmds.text(self.status_label, edit=True, label='ERROR: No materials selected')
            return False

        select_materials = [self.materials_filtered[index - 1] for index in user_selection]
        cmds.select(select_materials)
        cmds.hyperShade(objects='')
        selected_objects = cmds.ls(selection=True)
        cmds.text(self.status_label, edit=True, label=('INFO: ' + str(len(selected_objects)) + ' objects selected'))

    def assign_material_button_pressed(self, *args):
        """Assign material event handler. Assigns the selected material to the
        selected objects"""
        user_selection = cmds.textScrollList(self.material_list, query=True, selectItem=True)

        if not user_selection:
            cmds.text(self.status_label, edit=True, label='ERROR: No materials selected')
            return False

        if not cmds.ls(selection=True):
            cmds.text(self.status_label, edit=True, label='ERROR: No objects selected')
            return False

        cmds.hyperShade(assign=user_selection[0])
        cmds.text(self.status_label, edit=True, label='INFO: Material has been assigned')

    def select_material_button_pressed(self, *args):
        """Select material event handler. Selects the materials from the
        material list"""
        user_selection = cmds.textScrollList(self.material_list, query=True, selectItem=True)

        if not user_selection:
            cmds.text(self.status_label, edit=True, label='ERROR: No materials selected')
            return False

        cmds.select(user_selection, noExpand=True)
        cmds.text(self.status_label, edit=True, label='INFO: Material(s) selected')

    def graph_object_materials_button_pressed(self, *args):
        """Graph object materials button pressed. Graphs the materials of the
        selected objects"""
        shading_groups = []
        selected_objects = cmds.ls(selection=True, long=True)

        if not selected_objects:
            cmds.text(self.status_label, edit=True, label='ERROR: No objects selected')
            return False

        if not self.is_hypershade_open():
            cmds.text(self.status_label, edit=True, label='ERROR: Hypershade is not open')
            return False

        for current in selected_objects:
            shading_groups.extend(self.get_shading_group(current))

        shading_groups = set(shading_groups)
        shading_groups = list(shading_groups)
        cmds.select(shading_groups, noExpand=True)
        mel.eval('hyperShadePanelGraphCommand("hyperShadePanel1", "showUpstream");')
        cmds.select(selected_objects)

        cmds.text(self.status_label, edit=True, label='INFO: Assigned materials have been graphed')

    def setup_ui(self):
        """Creates the UI"""
        if cmds.window('materialListWindow', exists=True):
            cmds.deleteUI('materialListWindow')

        material_window = cmds.window('materialListWindow', title='Material List',
                                      width=265, height=625,
                                      minimizeButton=True, maximizeButton=False,
                                      sizeable=True, resizeToFitChildren=True)
        main_form = cmds.formLayout(numberOfDivisions=10, parent=material_window)
        self.material_frame = cmds.frameLayout(width=265, label='Materials: 0',
                                               collapsable=False, marginHeight=5, marginWidth=5)
        self.material_list = cmds.textScrollList(allowMultiSelection=True)
        self.filter_text = cmds.textFieldButtonGrp(label='Filter:', buttonLabel='Update List',
                                                   adjustableColumn=2, columnAlign=[1, 'left'],
                                                   columnWidth=[1, 35],
                                                   buttonCommand=self.refresh_material_list,
                                                   textChangedCommand=self.filter_text_changed)
        self.include_unused_checkbox = cmds.checkBox(label='Include Unused Materials',
                                                     changeCommand=self.refresh_material_list)
        self.include_sub_materials_checkbox = cmds.checkBox(label='Include Sub Materials',
                                                            changeCommand=self.refresh_material_list)
        cmds.setParent('..')  # end frameLayout

        shade_frame = cmds.frameLayout(width=265, label='Hypershade',
                                       collapsable=False, marginHeight=5, marginWidth=5)
        cmds.columnLayout(rowSpacing=5, adjustableColumn=True)
        cmds.rowColumnLayout(numberOfColumns=2, columnWidth=[1, 175], columnSpacing=[2, 10])
        cmds.button(width=85, label='Graph Selected Materials',
                    command=self.graph_material_button_pressed)
        cmds.radioCollection()
        self.graph_input_radio = cmds.radioButton(label='Input', select=True)
        cmds.text(label='')
        self.graph_output_radio = cmds.radioButton(label='Output')
        cmds.text(label='')
        cmds.radioButton(label='Both')
        cmds.setParent('..')  # end rowColumnLayout
        cmds.button(width=85, label='Clear Graph',
                    command=self.clear_graph_button_pressed)
        cmds.button(width=85, label='Open Hypershade',
                    command=self.open_hypershade_button_pressed)
        cmds.setParent('..')  # end columnLayout
        cmds.setParent('..')  # end frameLayout

        misc_frame = cmds.frameLayout(width=265, label='Misc',
                                      collapsable=False, marginHeight=5, marginWidth=5)
        cmds.columnLayout(rowSpacing=5, adjustableColumn=True)
        cmds.button(label='Select Objects By Material',
                    command=self.select_objects_by_material_button_pressed)
        cmds.button(label='Assign Material To Selected Objects',
                    command=self.assign_material_button_pressed)
        cmds.button(label='Select Material',
                    command=self.select_material_button_pressed)
        cmds.button(label='Graph Materials of Selected',
                    command=self.graph_object_materials_button_pressed)
        cmds.setParent('..')  # end columnLayout
        cmds.setParent('..')  # end frameLayout

        self.status_label = cmds.text(label='INFO: Standing by...', align='left')
        cmds.setParent('..')  # end formLayout

        cmds.formLayout(main_form, edit=True,
            attachForm=[(self.material_frame, 'top', 0),
                        (self.material_frame, 'left', 0),
                        (self.material_frame, 'right', 0),
                        (shade_frame, 'left', 0),
                        (shade_frame, 'right', 0),
                        (misc_frame, 'left', 0),
                        (misc_frame, 'right', 0),
                        (self.status_label, 'left', 5),
                        (self.status_label, 'right', 5),
                        (self.status_label, 'bottom', 5)],
            attachControl=[(self.material_frame, 'bottom', 5, shade_frame),
                           (shade_frame, 'bottom', 5, misc_frame),
                           (misc_frame, 'bottom', 5, self.status_label)])

        self.refresh_material_list()

        cmds.showWindow(material_window)


def main():
    """Main entry point. Opens the UI"""
    ml = MaterialList()
    ml.setup_ui()


if __name__ == "__main__":
    main()
